/* corr.c --- translate correlation names to table names */
/*
		   Copyright 1994 In Touch Ministries
		   Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include "ferret.h"
# include "store.h"
# ifdef VMS
#	include "y_tab.h"
# else
#	include "y.tab.h"
# endif

typedef struct Link {
    struct  Link    *next;
    string_t        table_name;
    string_t        corr_name;
} link_t;

static link_t *Head = NULL;
static long Ntables = 0;

static link_t *lookup (char *);
static int find_corr (enode_t *);
static int chg_corr (enode_t *);
static void free_corr (void);

extern char *progname;

/* trans_corr --- translate correlation names to table names */

void
trans_corr ()
{
    int i;
    enode_t *e;

    if (!DML)
	return;

    /* for each DML */
    for (i = 0; i < DML->used; i++) {
	Head = NULL;
	Ntables = 0;
	e = DML->elist[i];
	(void) walk (e, find_corr);     /* find them */
	(void) walk (e, chg_corr);      /* change them */
	free_corr ();                   /* free the memory */
    }
}

/* find_corr --- save table/corr_names if op == O_ALIAS */

 int
find_corr (enode_t *e)
{
    if (e->op == O_ALIAS) {
	link_t *l;

if (e->e.alias.table_name <= 200)
    fprintf (stderr, "bad table pointer\n");
if (e->e.alias.corr_name <= 200)
    fprintf (stderr, "bad corr pointer\n");
	/* only save it once */
	for (l = Head; l; l = l->next) {
	    if (l->table_name && e->e.alias.table_name &&
	      strcmp (l->table_name, e->e.alias.table_name) == 0 &&
	      l->corr_name && e->e.alias.corr_name &&
	      strcmp (l->corr_name, e->e.alias.corr_name) == 0)
		return (WALK);
	}
		
	l = emalloc (sizeof (link_t));
	l->table_name = e->e.alias.table_name;
	l->corr_name = e->e.alias.corr_name;
	l->next = Head;
	Head = l;
	++Ntables;
    }
    return (WALK);
}

/* free_corr --- free up the memory used by find_corr () */

static void
free_corr ()
{
    link_t *l, *n;

    for (l = Head; l; l = n) {
	n = l->next;
	free (l);
    }
}

/* chg_corr --- change all correlations to table names */

 int
chg_corr (enode_t *e)
{
    if (e->op == O_COL) {
	link_t *l;
	char buf[32+32+2];  /* table+column+'.'+'\0' */
	char *p;
	int newlen;

	strcpy (buf, e->e.lit_name);
	p = strchr (buf, '.');
	if (p == NULL && Ntables != 1)
	    return;
	
	if (p)
	    *p++ = 0;
	else
	    p = buf;
	/* now, buf holds the terminated corr name, and
	** p points to the column name
	*/
	if ((l = lookup (buf)) == NULL)
	    return;     /* Hmm, a corr_name I can't find */
	
	if (p != buf && strcmp (l->table_name, buf) == 0)
	    return;     /* already IS table_name */
	
	newlen = strlen (l->table_name) + strlen (p) + 2;
	e->e.lit_name = erealloc (e->e.lit_name, newlen);
	sprintf (e->e.lit_name, "%s.%s", l->table_name, p);
    }
    return (WALK);
}

/* lookup --- find this corr_name (or table_name) */

static link_t *
lookup (string_t name)
{
    link_t *l;

    if (Ntables == 1)
	return (Head);

    if (name == NULL)
	return (NULL);
	
    for (l = Head; l; l = l->next)
	if ((l->corr_name && strcmp (l->corr_name, name)) == 0 ||
	  (l->table_name && strcmp (l->table_name, name)) == 0) {
	    return (l);
	}

    return (NULL);
}
