/* errsum.c --- error summary reports */
/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <errno.h>
# include "ferret.h"
# include "store.h"
# ifdef VMS
#	include "y_tab.h"
# else
#	include "y.tab.h"
# endif

# define YES        1
# define NO         0

typedef struct List {
    struct      List    *next;
    string_t    name;
} list_t;

static list_t *From_list = NULL;
static list_t *Where_list = NULL;
static FILE *fp;

static char *Filename;
static int Lineno;

static void cart_prod ();
static void join_or ();
static void ill_join ();
static int isjoin (enode_t *);

extern char *progname;

/* error_sum --- produce the error summary report */

void
errsum (string_t outbase, string_t ext)
{
    char fname[256];
    extern int Pagelength;

    if (!DML)
	return;

    sprintf (fname, "%s%c%s", outbase, EXTSEP, ext);
    if ((fp = fopen (fname, "w")) == NULL) {
	fprintf (stderr, "%s: can't open %s for writing (%s)\n",
	  progname, fname, strerror (errno));
	return;
    }

    pg_open (fp, Pagelength, 80, 5, 5);

    pg_head (1, "|ERROR SUMMARY|%D", '|');

    pg_head (2, "|%T", '|');

    cart_prod ();   /* look for cartesian products */

    join_or ();     /* look for joins with an OR parent node */

    ill_join ();    /* look for illogical joins */

    pg_close ();

    fclose (fp);
}

static void
cart_prod ()
{
    int i, ntables;
    list_t *f, *l;
    static int cart_look (enode_t *);
    static int where_look (string_t);

    for (i = 0; i < DML->used; i++) {
	Filename = DML->elist[i]->e.dml->fname;
	Lineno = DML->elist[i]->e.dml->lineno;
	(void) walk (DML->elist[i], cart_look);
	
	ntables = 0;
	for (f = From_list; f; f = f->next)
	    ++ntables;
	
	/* cart-prod only possible if at least two tables involved */
	if (ntables > 1)
	    for (f = From_list; f; f = f->next)
		if (where_look (f->name) == NO)
		    pg_printf ("%s @ %d: possible cartesian product on table %s\n",
		      Filename, Lineno, f->name);
		
	for (f = From_list; f; f = l) {
	    l = f->next;
	    free (f->name);
	    free (f);
	}
	From_list = NULL;
	
	for (f = Where_list; f; f = l) {
	    l = f->next;
	    free (f->name);
	    free (f);
	}
	Where_list = NULL;
    }
}

/* cart_look --- walk through the from and where clauses */

 int
cart_look (enode_t *e)
{
    list_t *f;
    static int from_walk (enode_t *);
    static int where_walk (enode_t *);

    if (e->op != O_SUBSEL)
	return (WALK);

    /* if the statement was "WHERE :variable", then we can't tell
    ** if there may be a cart-prod or not, so we punt
    */
    if (e->e.subsel->where_clause && e->e.subsel->where_clause->op == O_LIT)
	return (WALK);
	
    (void) walk (e->e.subsel->from_clause, from_walk);
    (void) walk (e->e.subsel->where_clause, where_walk);
    return (WALK);
}

/* where_look --- step through the Where_list */

static int
where_look (string_t name)
{
    list_t *w;

    for (w = Where_list; w; w = w->next)
	if (strcmp (w->name, name) == 0)
	    return (YES);

    return (NO);
}

/* from_walk --- look through the from nodes for table names */

static int
from_walk (enode_t *e)
{
    list_t *p;

    if (e->op == O_ALIAS) {
	p = emalloc (sizeof (list_t));
	p->name = strdup (e->e.alias.table_name);
	p->next = From_list;
	From_list = p;
    }
    return (WALK);
}

/* where_walk --- look through the where nodes for table names */

static int
where_walk (enode_t *e)
{
    list_t *p;
    char buf[128], *d;

    if (e->op == O_COL) {
	p = emalloc (sizeof (list_t));
	strcpy (buf, e->e.lit_name);
	if (d = strchr (buf, '.'))
	    *d = 0;
	p->name = strdup (buf);
	p->next = Where_list;
	Where_list = p;
    }
    return (WALK);
}


/* join_or --- look for joins with a parent OR node */

static void
join_or ()
{
    int i;
    static int or_look (enode_t *);

    for (i = 0; i < DML->used; i++) {
	Filename = DML->elist[i]->e.dml->fname;
	Lineno = DML->elist[i]->e.dml->lineno;
	(void) walk (DML->elist[i], or_look);
    }
}

/* or_look --- look for OR nodes */

static int
or_look (enode_t *e)
{
    if (e->op == OR) {
	if (isjoin (e->e.child.left) || isjoin (e->e.child.right))
	    pg_printf ("%s @ %d: Join with OR\n", Filename, Lineno);
    }
    return (WALK);
}

/* isjoin --- is this a join? */

static int
isjoin (enode_t *e)
{
    enode_t *l, *r;

    if (e->op == EQ) {
	l = e->e.child.left;
	r = e->e.child.right;
	if (l->op == O_COL && r->op == O_COL)
	    return (1);
    }
    return (0);
}

/* ill_join --- look for illogical joins */

static void
ill_join ()
{
    int i;
    static int ill_look (enode_t *);

    for (i = 0; i < DML->used; i++) {
	Filename = DML->elist[i]->e.dml->fname;
	Lineno = DML->elist[i]->e.dml->lineno;
	(void) walk (DML->elist[i], ill_look);
    }
}

/* ill_look --- look for EQ nodes */

static int
ill_look (enode_t *e)
{
    enode_t *l, *r;

    if (e->op == EQ) {
	l = e->e.child.left;
	r = e->e.child.right;
	if (l->op == O_COL && r->op == O_COL &&
	  strcmp (l->e.lit_name, r->e.lit_name) == 0)
	    pg_printf ("%s @ %d: %s joined with %s\n", Filename,
	      Lineno, l->e.lit_name, r->e.lit_name);
    }
    return (WALK);
}
