/* ferret.c --- executive */
/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <errno.h>
# include "ferret.h"

# define DEFPAGELENGTH  66

char *progname;                 /* progname name for error messages */
char *Fname;                    /* current file name */
FILE *Fin;                      /* current input file file pointer */
int lineno = 1;                 /* what input line we're on */
char *Outbase = "ferret";       /* base name for output files (ferret.wcs) */
int Dosyntax = 0;               /* syntax errors off by default */
char *Dbname = NULL;            /* dbname for optimize */
int Pagelength = DEFPAGELENGTH; /* default page length */
static char *Listfile = NULL;   /* file of file names to read */
static int Verbose = 0;         /* print file names when processing */
int Filetype = 0;				/* ABF, C */

static report_t rep_tab[] = {
    { syntax,     "syn" },  /* syntax error summary */
    { where_use,  "wcs" },  /* where clause summary */
    { cross_ref,  "xrf" },  /* table/column cross ref */
    { optimize,   "opt" },  /* optimizedb paramater file */
    { errsum,     "err" },  /* error summary */
    { join_ref,   "jrf" },  /* join column cross ref */
    { warnsum,    "wrn" },  /* warning summary */
    { NULL,       NULL  }
};

static void process (string_t);
static void dolist (void);

extern int yyparse (void);
extern void set_lex_skim (void);
extern void trans_corr (void);
extern void print_dml (void);

main (int argc, char **argv)
{
    int i, c, err = 0;
    int printdml = 0;
    char *startmem;
    report_t *rp;
    extern int yydebug, optind;
    extern char *optarg;
    extern void print_lex_count (void);
	extern char *sbrk (int);

    startmem = sbrk (0);    /* beginning of heap */
    progname = argv[0];

    yydebug = 0;
    printdml = 0;
    while ((c = getopt (argc, argv, "b:sypf:vd:l:")) != EOF) {
	switch (c) {
	case 'b':
	    Outbase = optarg;
	    break;
	case 's':
	    Dosyntax = 1;
	    break;
	case 'y':
	    yydebug = 1;
	    break;
	case 'p':
	    printdml = 1;
	    break;
	case 'f':
	    Listfile = optarg;
	    break;
	case 'v':
	    Verbose = 1;
	    break;
	case 'd':
	    Dbname = optarg;
	    break;
	case 'l':
	    Pagelength = atoi (optarg);
	    if (Pagelength <= 10) {
		fprintf (stderr, "%s: unreasonable page length: %d.  Setting to %d\n",
		  progname, Pagelength, DEFPAGELENGTH);
		Pagelength = DEFPAGELENGTH;
	    }
	    break;
	default:
	    ++err;
	    break;
	}
    }
    if (err || (optind >= argc && Listfile == NULL)) {
	fprintf (stderr, "Usage: %s [-s] [-y] [-p] [-v]\n", progname);
	fprintf (stderr, "       [-d dbname] [-b basename] [-f listfile]\n");
	fprintf (stderr, "       [-l length] [file] ...\n");
	exit (1);
    }

    if (Listfile != NULL)
	dolist ();

    for (i = optind; i < argc; i++)
# ifdef VMS
	if (strchr (argv[i], '*'))	/* wildcard */
	    expand (argv[i]);
	else
# endif /* VMS */
	    process (argv[i]);

    trans_corr ();  /* translate corr names to table names */

    if (printdml)
	print_dml ();   /* print all the DMLs */

    /* print all the reports */
    for (rp = rep_tab; rp->report; ++rp) {
	if (Verbose)
	    fprintf (stderr, "%s: generating %s report\n", progname, rp->ext);
	(*(rp->report)) (Outbase, rp->ext);
    }
	
    if (Verbose)
	fprintf (stderr, "%s: allocated %ld bytes\n", progname,
	  (char *) sbrk (0) - startmem);
	
    exit (0);

    /* NOTREACHED */
}

/* process --- process one input file */

static void
process (string_t input)
{
    char *p;

    Fname = input;
    if ((Fin = fopen (Fname, "r")) == NULL) {
	fprintf (stderr, "%s: can't open %s for reading (%s)\n",
	  progname, Fname, strerror (errno));
	return;
    }

    if (p = strchr (input, ']'))
	Fname = p+1;	/* skip VMS "disk:[dir]" part */

    if (Verbose)
	fprintf (stderr, "%s\n", Fname);
	
    if (p = strchr (Fname, '.')) {
	++p;
	if (strcmp (p, "sc") == 0)
		Filetype = CSRC;
	else
		Filetype = ABFSRC;
    }

    lineno = 1;

    while (!feof (Fin)) {
	(void) yyparse ();
	set_lex_skim ();
    }

    fclose (Fin);
}

/* dolist --- read and process the files in Listfile */

static void
dolist ()
{
    FILE *fp;
    char *p;
    static char line[256];

    if ((fp = fopen (Listfile, "r")) == NULL) {
	fprintf (stderr, "%s: can't open %s for reading (%s)\n",
	  progname, Listfile, strerror (errno));
	return;
    }

    while (fgets (line, sizeof (line), fp) != NULL) {
	if (p = strchr (line, '\n'))
	    *p = 0;
	
	process (line);
    }

    fclose (fp);
}

/* getopt --- get the options from argv */

int optind;
char *optarg;

int
getopt (int ac, char **av, char *optstr)
{
	static int first = 1;
	static int sac;
	static char **sav;
	char *p;

	if (first) {
		first = 0;
		sac = ac;
		sav = av;
		optind = 0;
	}

	++optind;
	if (*sav[optind] == '-') {
		if ((p = strchr (optstr, sav[optind][1]))) {
			if (p[1] == ':')
				optarg = &sav[optind][2];
			return (sav[optind][1]);
		}
		return ('?');
	}
	return (EOF);
}

# ifdef VMS

/* expand --- expand a filespec wild card */

# include <rmsdef.h>
# include <descrip.h>

static int
expand (char *wild)
{
    char fname[80], *p;
    $DESCRIPTOR (in_spec_d, wild);
    $DESCRIPTOR (out_spec_d, fname);
    int cond_value, context;

    in_spec_d.dsc$w_length = strlen (wild);
    context = 0;
    while ((cond_value = lib$find_file (&in_spec_d, &out_spec_d, &context))
      != RMS$_NMF) {
	if ((cond_value&1) != 1)
	    lib$stop (cond_value);
	if ((p = strchr (fname, ' ')))
	    *p = 0;
	process (fname);
    }
    lib$find_file_end (&context);
}

# endif /* VMS */
