/* join_ref.c --- join cross reference */
/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <errno.h>
# include "ferret.h"
# include "store.h"
# ifdef VMS
#	include "y_tab.h"
# else
#	include "y.tab.h"
# endif

/* type for table/column cross reference */

typedef struct {
    char    *l_table_name;  /* left table */
    char    *l_column_name; /* left column */
    int     op;             /* operator (=, <=, !=) */
    char    *r_table_name;  /* right table */
    char    *r_column_name; /* right column */
    char    *file_name;     /* source file name */
    long    lineno;         /* source line number */
} join_ref_t;

typedef struct Count {
    struct  Count   *next;
    char    *table;
    int     count;
} count_t;

static join_ref_t **Join_ref;       /* array of -> join_ref_ts */
static int Njoin_ref = 0;           /* how many join_ref_ts */
static int MAXjoin_ref = 0;         /* how many allocated */

static count_t *Count = NULL;

static char *file_name;
static long lineno;

static void jr_new (enode_t *, int, enode_t *);
static void jr_tab_count (char *);
static int  jr_tab_total (char *);
static void jr_swap ();
static int  jr_store (enode_t *);
static int  jr_walk (enode_t *);
static int  jr_cmp (join_ref_t **, join_ref_t **);
static char *doop (int);

extern char *progname;

void
join_ref (string_t outbase, string_t ext)
{
    int i;
    join_ref_t *j;
    char fname[256];
    FILE *fp;
    count_t *p, *n;
    extern int Pagelength;

    if (!DML)
	return;

    /* run through the where clauses, and store tables and columns */
    for (i = 0; i < DML->used; i++)
	(void) walk (DML->elist[i], jr_walk);

    jr_swap ();

    qsort (Join_ref, Njoin_ref, sizeof (join_ref_t *), jr_cmp);

    sprintf (fname, "%s%c%s", outbase, EXTSEP, ext);
    if ((fp = fopen (fname, "w")) == NULL) {
	fprintf (stderr, "%s: can't open %s for writing (%s)\n",
	  progname, fname, strerror (errno));
	return;
    }

    if (Njoin_ref > 0) {
	char line[128];
	
	pg_open (fp, Pagelength, 115, 5, 5);
	pg_head (1, "|TABLE/COLUMN JOIN CROSS REFERENCE|%D", '|');
	pg_head (2, "|%T", '|');
	sprintf (line, "%-20.20s %20.20s  %2.2s  %-20.20s %-20.20s %-20.20s %4s|",
	  "Table",
	  "Column",
	  "OP",
	  "Column",
	  "Table",
	  "File",
	  "Line"
	  );
	pg_head (3, line, '|');
	pg_head (4,
	  "------------------------------------------------------------------------------------------------------------------|",
	  '|');
    }

    for (i = 0; i < Njoin_ref; i++) {
	j = Join_ref[i];
	pg_printf ("%-20.20s %20.20s  %2.2s  %-20.20s %-20.20s %-20.20s %4ld\n",
	  j->l_table_name,
	  j->l_column_name,
	  doop (j->op),
	  j->r_column_name,
	  j->r_table_name,
	  j->file_name,
	  j->lineno
	  );
	free (j->l_table_name);
	free (j->l_column_name);
	free (j->r_table_name);
	free (j->r_column_name);
	free (j->file_name);
	free (j);
    }
    if (Join_ref) {
	free (Join_ref);
	Join_ref = NULL;
	MAXjoin_ref = 0;
	Njoin_ref = 0;
    }

    for (p = Count; p; p = n) {
	n = p->next;
	free (p);
    }
    Count = NULL;

    pg_close ();

    fclose (fp);
}

/* jr_walk --- called for each node of each DML */

static int
jr_walk (enode_t *e)
{

    if (e->op == O_DML) {
	file_name = e->e.dml->fname;
	lineno = e->e.dml->lineno;
	(void) walk (e->e.dml->where_clause, jr_store);
    }
    else if (e->op == O_SUBSEL) {
	(void) walk (e->e.subsel->where_clause, jr_store);
    }
    return (WALK);
}

/* jr_store --- for certain operators, store 'em */

static int
jr_store (enode_t *e)
{
    enode_t *r, *l;

    if (e->op == EQ ||
      e->op == NE ||
      e->op == LT ||
      e->op == LE ||
      e->op == GT ||
      e->op == GE) {
	l = e->e.child.left;
	r = e->e.child.right;
	if (l->op == O_COL && r->op == O_COL)
	    jr_new (l, e->op, r);
    }
    return (WALK);
}

/* jr_new --- return a new join_ref_t */

static void
jr_new (enode_t *r, int op, enode_t *l)
{
    int i;
    char *p;
    char name[128];
	
    if (Njoin_ref >= MAXjoin_ref) {
	MAXjoin_ref += 10;
	if (Join_ref == NULL)
	    Join_ref = emalloc (MAXjoin_ref * sizeof (Join_ref[0]));
	else
	    Join_ref = erealloc (Join_ref,
	      MAXjoin_ref * sizeof (Join_ref[0]));
    }
	
    i = Njoin_ref++;
    Join_ref[i] = emalloc (sizeof (join_ref_t));
    memset (Join_ref[i], 0, sizeof (join_ref_t));
    strcpy (name, l->e.lit_name);
    if ((p = strchr (name, '.'))) {
	*p = 0;
	Join_ref[i]->l_table_name = strdup (name);
	Join_ref[i]->l_column_name = strdup (p+1);
    }
    else {
	Join_ref[i]->l_column_name = strdup (name);
	Join_ref[i]->l_table_name = strdup ("*** AMBIGUOUS ***");
    }
    strcpy (name, r->e.lit_name);
    if ((p = strchr (name, '.'))) {
	*p = 0;
	Join_ref[i]->r_table_name = strdup (name);
	Join_ref[i]->r_column_name = strdup (p+1);
    }
    else {
	Join_ref[i]->r_column_name = strdup (name);
	Join_ref[i]->r_table_name = strdup ("*** AMBIGUOUS ***");
    }
    Join_ref[i]->op = op;
    Join_ref[i]->file_name = strdup (file_name);
    Join_ref[i]->lineno = lineno;
    jr_tab_count (Join_ref[i]->r_table_name);
    jr_tab_count (Join_ref[i]->l_table_name);
}

/* jr_cmp --- compare two elements for qsort */

static int
jr_cmp (join_ref_t **a, join_ref_t **b)
{
    int cmp;

    cmp = strcmp ((*a)->l_table_name, (*b)->l_table_name);
    if (cmp)
	return (cmp);
    cmp = strcmp ((*a)->l_column_name, (*b)->l_column_name);
    if (cmp)
	return (cmp);
    cmp = (*a)->op - (*b)->op;
    if (cmp)
	return (cmp);
    cmp = strcmp ((*a)->r_table_name, (*b)->r_table_name);
    if (cmp)
	return (cmp);
    cmp = strcmp ((*a)->r_column_name, (*b)->r_column_name);
    if (cmp)
	return (cmp);
    cmp = strcmp ((*a)->file_name, (*b)->file_name);
    if (cmp)
	return (cmp);
    return ((*a)->lineno - (*b)->lineno);
}

/* doop --- return a printable version of type */

static char *
doop (int op)
{
    switch (op) {
    case EQ:
	return ("==");
	break;
    case NE:
	return ("!=");
	break;
    case LT:
	return ("<");
	break;
    case LE:
	return ("<=");
	break;
    case GT:
	return (">");
	break;
    case GE:
	return (">=");
	break;
    default:
	return ("I don't know");
	break;
    }
}

/* jr_tab_count --- store the number of times this table is mentioned */

static void
jr_tab_count (char *table)
{
    count_t *pc;

    for (pc = Count; pc; pc = pc->next)
	if (strcmp (table, pc->table) == 0) {
	    ++(pc->count);
	    return;
	}

    pc = emalloc (sizeof (count_t));
    pc->table = table;
    pc->count = 1;
    pc->next = Count;
    Count = pc;
}

/* jr_tab_total --- return the total count for this table */

static int
jr_tab_total (char *table)
{
    count_t *pc;

    for (pc = Count; pc; pc = pc->next)
	if (strcmp (table, pc->table) == 0)
	    return (pc->count);
	
    return (-1);
}

/* jr_swap --- swap the table/column pairs, if necessary */

static void
jr_swap ()
{
    int i;
    join_ref_t *j;

    for (i = 0; i < Njoin_ref; i++) {
	j = Join_ref[i];
	if (jr_tab_total (j->l_table_name) <
	  jr_tab_total (j->r_table_name)) {
	    char *tmp_tab, *tmp_col;
	
	    tmp_tab = j->l_table_name;
	    tmp_col = j->l_column_name;
	
	    j->l_table_name = j->r_table_name;
	    j->l_column_name = j->r_column_name;
	
	    j->r_table_name = tmp_tab;
	    j->r_column_name = tmp_col;
	
	    /* must swap these operators, too */
	    if (j->op == GT)
		j->op = LT;
	    else if (j->op == GE)
		j->op = LE;
	    else if (j->op == LT)
		j->op = GT;
	    else if (j->op == GE)
		j->op = GE;
	}
    }
}
