/* parse.y --- input grammer for yacc */
/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

%{
# include <stdio.h>
# include "ferret.h"
# include "store.h"

# define YYDEBUG    1
%}

%union {
    string_t    lit;
    enode_t     *enode;
    elist_t     *elist;
    subsel_t    *sub;
    int         ord;
    fileloc_t   *floc;
}

%token      <lit>       ALL ANY AS ASC AT BETWEEN BY DESC DISTINCT
%token      <lit>       FROM GROUP HAVING IN INTO IS LIKE SQL_NULL OF
%token      <lit>       ORDER SET SOME TO UNION VALUES WHERE WITH
%token      <lit>       IDENT LITERAL ASGN UNION_ALL EXISTS
%token      <lit>       NOT_SQL_NULL QUAL
%token      <floc>      DELETE INSERT SELECT UPDATE
%nonassoc   <enode>     ESCAPE
%left       <enode>     OR
%left       <enode>     AND
%left       <enode>     GT GE LT LE EQ NE
%left       <enode>     '+' '-'
%left       <enode>     '*' '/'
%left       <enode>     UMINUS UPLUS NOT
%right      <enode>     EXPON
%left       <enode>     CPTR '.'
%type	    <lit>	array_spec cptr
%type       <enode>     col_or_lit subselect escape_clause
%type       <enode>     where_clause having_clause
%type       <enode>     result table_corr insert_values
%type       <enode>     boolean expr asgn sub_sel_union qual_asgn
%type       <elist>     expr_list expr_list1 ins_column_list
%type       <elist>     column_list asgn_list table_list order_clause
%type       <elist>     order_list from_clause group_clause result_list
%type       <elist>     col_or_lit_list qual_asgn_list
%type       <enode>     dml insert delete update select literal
%type       <enode>     any_or_all end expr_or_asgn into_clause
%type       <ord>       dir all_dist cmp
%%
dml:                  /* nothing */         { $$ = NULL; }
		    | dml insert  end       { store ($2); YYACCEPT; }
		    | dml delete  end       { store ($2); YYACCEPT; }
		    | dml update  end       { store ($2); YYACCEPT; }
		    | dml select  end       { store ($2); YYACCEPT; }
		    | dml error             { YYABORT; }
		    ;

end:                  /* nothing */ { $$ = NULL; }
		    | ';'           { $$ = NULL; }
		    | IDENT         { $$ = NULL; }
		    | ')'           { $$ = NULL; }
		    ;

insert:               INSERT INTO table_list ins_column_list insert_values
			{ $$ = new_ins ($1, $3, $4, $5); }
			    
		    ;

insert_values:        VALUES '(' expr_list1 ')' { $$ = lton ($3); }
		    | subselect                 { $$ = $1; }
		    ;

expr_list1:           expr                  { $$ = new_elist (NULL, $1); }
		    | expr_list1 ',' expr   { $$ = new_elist ($1, $3); }
		    ;

ins_column_list:      /* nothing */         { $$ = NULL; }
		    | '(' column_list ')'   { $$ = $2; }
		    ;
		    
column_list:          IDENT                 { $$ = new_elist (NULL, col ($1)); }
		    | column_list ',' IDENT { $$ = new_elist ($1, col ($3)); }
		    ;

delete:               DELETE FROM table_list where_clause
		      { $$ = new_del ($1, $3, $4); }
		    ;

where_clause:         /* nothing */     { $$ = NULL; }
		    | WHERE boolean     { $$ = new_enode (O_WHERE, $2, NULL); }
		    | WHERE literal     { $$ = new_enode (O_WHERE, $2, NULL); }
		    ;

update:               UPDATE table_list from_clause SET asgn_list where_clause
		      { $$ = new_upd ($1, $2, $3, $5, $6); }
		    ;

asgn_list:            asgn                  { $$ = new_elist (NULL, $1); }
		    | asgn_list ',' asgn    { $$ = new_elist ($1, $3); }
		    ;

asgn:                 IDENT EQ expr     { $$ = new_enode (ASGN, col ($1), $3); }
		    | literal EQ expr   { $$ = new_enode (ASGN, $1, $3); }
		    ;
		    
select:               sub_sel_union order_clause
		      { $$ = new_sel ($1, $2); }
		    ;

order_clause:         /* nothing */         { $$ = NULL; }
		    | ORDER BY order_list   { $$ = $3; }
		    ;

order_list:           col_or_lit dir    { $$ = new_order (NULL, $1, $2); }
		    | order_list ',' col_or_lit dir {
		      $$ = new_order ($1, $3, $4); }
		    ;

col_or_lit:           IDENT     { $$ = col ($1); }
		    | literal   { $$ = $1; }
		    ;

col_or_lit_list:      col_or_lit    { $$ = new_elist (NULL, $1); }
		    | col_or_lit_list ',' col_or_lit {
		      $$ = new_elist ($1, $3); }
		    ;

dir:                  /* nothing */     { $$ = ASC; }
		    | ASC               { $$ = ASC; }
		    | DESC              { $$ = DESC; }
		    ;

sub_sel_union:        subselect                     { $$ = $1; }
		    | subselect UNION subselect    
		      { $$ = new_enode (UNION, $1, $3); }
		    | subselect UNION ALL subselect
		      { $$ = new_enode (UNION_ALL, $1, $4); }
		    ;

subselect:            SELECT all_dist result_list into_clause
			  from_clause where_clause group_clause
			  having_clause
		      { $$ = new_subsel ($1, $2, $3, $4, $5, $6, $7, $8); }
		    ;

all_dist:             /* nothing */     { $$ = ALL; }
		    | ALL               { $$ = ALL; }
		    | DISTINCT          { $$ = DISTINCT; }
		    ;

result_list:          result                    { $$ = new_elist (NULL, $1); }
		    | '*'                       { $$ = new_elist (NULL, lit ("*")); }
		    | result_list ',' result    { $$ = new_elist ($1, $3); }
		    ;

result:               expr              { $$ = $1; }
		    | expr AS IDENT     { $$ = new_enode (ASGN, lit ($3), $1); }
		    | IDENT EQ expr     { $$ = new_enode (ASGN, lit ($1), $3); }
		    | literal EQ expr   { $$ = new_enode (ASGN, $1, $3); }
		    ;
		    
into_clause:          /* nothing */     { $$ = NULL; }
		    | INTO col_or_lit_list {
		      $$ = new_enode (INTO, lton ($2), NULL); }
		    ;

from_clause:          /* nothing */     { $$ = NULL; }
		    | FROM table_list   { $$ = $2; }
		    ;

table_list:           table_corr                { $$ = new_elist (NULL, $1); }
		    | table_list ',' table_corr { $$ = new_elist ($1, $3); }
		    ;

table_corr:           IDENT         { $$ = new_alias ($1, NULL); }
		    | IDENT   IDENT { $$ = new_alias ($1, $2); }
		    | LITERAL       { $$ = new_alias ($1, NULL); }
		    | LITERAL IDENT { $$ = new_alias ($1, $2); }
		    ;

group_clause:         /* nothing */         { $$ = NULL; }
		    | GROUP BY column_list  { $$ = $3; }
		    ;

having_clause:        /* nothing */         { $$ = NULL; }
		    | HAVING boolean        { $$ = $2; }
		    ;

boolean:              expr GT expr          { $$ = new_enode (GT, $1, $3); }
		    | expr GE expr          { $$ = new_enode (GE, $1, $3); }
		    | expr LT expr          { $$ = new_enode (LT, $1, $3); }
		    | expr LE expr          { $$ = new_enode (LE, $1, $3); }
		    | expr EQ expr          { $$ = new_enode (EQ, $1, $3); }
		    | expr NE expr          { $$ = new_enode (NE, $1, $3); }
		    | boolean AND boolean   { $$ = new_enode (AND, $1, $3); }
		    | boolean OR  boolean   { $$ = new_enode (OR, $1, $3); }
		    | NOT boolean           { $$ = new_enode (NOT, $2, NULL); }
		    | '(' boolean ')'       { $$ = $2; }
		    | expr LIKE literal escape_clause {
		      $$ = new_like ($1, 1, $3, $4); }
		    | expr NOT LIKE literal escape_clause {
		      $$ = new_like ($1, 0, $4, $5); }
		    | expr BETWEEN expr AND expr {
		      $$ = new_between ($1, 1, $3, $5); }
		    | expr NOT BETWEEN expr AND expr {
		      $$ = new_between ($1, 0, $4, $6); }
		    | expr IN '(' col_or_lit_list ')' {
		      $$ = new_in ($1, 1, lton ($4)); }
		    | expr NOT IN '(' col_or_lit_list ')' {
		      $$ = new_in ($1, 0, lton ($5)); }
		    | expr IN '(' subselect ')' {
		      $$ = new_in ($1, 1, $4); }
		    | expr NOT IN '(' subselect ')' {
		      $$ = new_in ($1, 0, $5); }
		    | any_or_all '(' subselect ')' {
		      $1->e.any_all.subsel = $3; $$ = $1; }
		    | EXISTS '(' subselect ')' {
		      $$ = new_enode (EXISTS, $3, NULL); }
		    | expr IS SQL_NULL {
		      $$ = new_enode (SQL_NULL, $1, NULL); }
		    | expr IS NOT SQL_NULL {
		      $$ = new_enode (NOT_SQL_NULL, $1, NULL); }
		    | QUAL '(' qual_asgn_list ')' { $$ = bltin ($1, $3); }
		    ;

qual_asgn_list:       /* nothing */     { $$ = NULL; }
		    | qual_asgn         { $$ = new_elist (NULL, $1); }
		    | qual_asgn_list ',' qual_asgn  { $$ = new_elist ($1, $3); }
		    ;

qual_asgn:            IDENT EQ IDENT    { $$ = new_enode (ASGN, col ($1), lit ($3)); }
		    ;

any_or_all:           cmp ANY { $$ = new_any ($1); }
		    | cmp ALL { $$ = new_all ($1); }
		    ;

cmp:                  EQ    { $$ = EQ; }
		    | NE    { $$ = NE; }
		    | LT    { $$ = LT; }
		    | LE    { $$ = LE; }
		    | GT    { $$ = GT; }
		    | GE    { $$ = GE; }
		    ;

escape_clause:        /* nothing */     { $$ = NULL; }
		    | ESCAPE literal    { $$ = new_enode (ESCAPE, $2, NULL); }
		    ;

expr:                 literal           { $$ = $1; }
		    | IDENT             { $$ = col ($1); }
		    | SQL_NULL          { $$ = lit ($1); }
		    | IDENT '(' expr_list ')'   { $$ = bltin ($1, $3); }
		    | '(' expr ')'      { $$ = $2; }
		    | '(' subselect ')' { $$ = $2; }
		    | expr '+' expr     { $$ = new_enode ('+', $1, $3); }
		    | expr '-' expr     { $$ = new_enode ('-', $1, $3); }
		    | expr '*' expr     { $$ = new_enode ('*', $1, $3); }
		    | expr '/' expr     { $$ = new_enode ('/', $1, $3); }
		    | '-' expr  %prec UMINUS {
		      $$ = new_enode (UMINUS, $2, NULL); }
		    | '+' expr  %prec UPLUS  { $$ = $2; }
		    | expr EXPON expr   {
		      $$ = new_enode (EXPON, $1, $3); }
		    ;

expr_list:            /* nothing */         { $$ = NULL; }
		    | expr_or_asgn          { $$ = new_elist (NULL, $1); }
		    | '*'                   { $$ = new_elist (NULL, lit ("*")); }
		    | expr_list ',' expr_or_asgn    { $$ = new_elist ($1, $3); }
		    ;

expr_or_asgn:         expr          { $$ = $1; }
		    | asgn          { $$ = $1; }
		    ;

literal:              LITERAL {
			    $$ = new_lit (NULL, $1, NULL, NULL); }
		    | LITERAL array_spec {
			    $$ = new_lit (NULL, $1, $2, NULL); }
		    | literal cptr LITERAL {
			    $$ = new_lit ($1, $3, $2, NULL); }
		    | literal cptr IDENT {
			    $$ = new_lit ($1, $3, $2, NULL); }
		    ;

cptr:		      CPTR          { $$ = "->"; }
		    | '.'           { $$ = "."; }
		    ;

array_spec:           '[' expr ']'      { $$ = $2; }
		    | '[' ']'           { $$ = "[]"; }
		    ;
%%
