/* store.c --- dml management routines */
/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include "ferret.h"
# include "store.h"
# ifdef VMS
#	include "y_tab.h"
# else
#	include "y.tab.h"
# endif

elist_t *DML = NULL;        /* holds all dml statements we find */
syntax_t *Syntax_l = NULL;  /* lines with syntax errors */

static void find_floc (enode_t *, long *, char **);
static int floc_visit (enode_t *);
static int print_join (enode_t *);
void resize_elist (void);

/* store --- store this top level dml for later */

void
store (enode_t *e)
{
    DML = new_elist (DML, e);
}

/* new_enode --- make a new enode */

enode_t *
new_enode (int op, enode_t *left, enode_t *right)
{
    enode_t *e;

    e = (enode_t *) emalloc (sizeof (enode_t));
    e->op = op;
    e->e.child.left = left;
    e->e.child.right = right;
    return (e);
}

/* new_lit --- build a new literal */

enode_t *
new_lit (enode_t *e, string_t lit, string_t cptr, string_t array_spec)
{
    char buf[128];
    int newlen;

    if (e == NULL) {
	e = emalloc (sizeof (enode_t));
	e->op = O_LIT;
	e->e.lit_name = lit;
    }

    *buf = 0;

    if (array_spec) {
	char tmp[128];
	if (strcmp (array_spec, "[]") != 0) {
	    strcpy (tmp, "[");
	    strcat (tmp, array_spec);
	    strcat (tmp, "]");
	}
	strcat (buf, array_spec);
    }
    if (cptr)
	strcat (buf, cptr);
	
    newlen = strlen (e->e.lit_name) + strlen (buf) + 1;
    e->e.lit_name = erealloc (e->e.lit_name, newlen);
    strcat (e->e.lit_name, buf);
    return (e);
}

/* bltin --- make a new built-in function def */

enode_t *
bltin (string_t func_name, elist_t *arg_list)
{
    enode_t *e;

    e = (enode_t *) emalloc (sizeof (enode_t));
    e->op = O_FUNC;
    e->e.func.func_name = func_name;
    e->e.func.arg_list = lton (arg_list);
    return (e);
}

/* lit --- make a new literal */

enode_t *
lit (string_t lit_name)
{
    enode_t *e;

    e = (enode_t *) emalloc (sizeof (enode_t));
    e->op = O_LIT;
    e->e.lit_name = lit_name;
    return (e);
}

/* col --- make a new column name */

enode_t *
col (string_t col_name)
{
    enode_t *e;

    e = (enode_t *) emalloc (sizeof (enode_t));
    e->op = O_COL;
    e->e.lit_name = col_name;
    locase (col_name);
    return (e);
}

/* cptr --- handle a LITERAL "->" IDENT expr */

enode_t *
cptr (string_t lit, string_t id)
{
    char buf[128];
    enode_t *e;

    e = emalloc (sizeof (enode_t));
    e->op = O_LIT;
    sprintf (buf, "%s->%s", lit, id);
    e->e.lit_name = emalloc (strlen (buf)+1);
    strcpy (e->e.lit_name, buf);
    return (e);
}

/* new_ins --- make a new insert enode */

enode_t *
new_ins (fileloc_t *floc, elist_t *table_list, elist_t *col_list, enode_t *ins_val)
{
    dml_t *d;
    enode_t *e;
    void mod_col_name (elist_t *, elist_t *);

    d = (dml_t *) emalloc (sizeof (dml_t));
    d->type = INSERT;
    d->lineno = floc->lineno;
    d->fname = floc->fname;
    free (floc);

    mod_col_name (table_list, col_list);
    d->table_list = lton (table_list);
    d->column_list = lton (col_list);
    d->where_clause = NULL;
    d->ins_val = ins_val;
    d->asgn_list = NULL;
    d->order_clause = NULL;
    d->sub_sel = NULL;
    d->from_clause = NULL;
    e = (enode_t *) emalloc (sizeof (enode_t));
    e->op = O_DML;
    e->e.dml = d;
    return (e);
}

static char *Table_name;

/* mod_col_name --- prepend the table name on each column */

static void
mod_col_name (elist_t *table_list, elist_t *col_list)
{
    int i;
    static int add_table_name (enode_t *);

    /* only INSERTs and UPDATEs use this, so only one table in the list */
    Table_name = table_list->elist[0]->e.lit_name;

    if (col_list == NULL)
	return;

    for (i = 0; i < col_list->used; i++)
	(void) walk (col_list->elist[i], add_table_name);
}

/* add_table_name --- prepend the table name on one enode */

static int
add_table_name (enode_t *e)
{
    char newname[24+32+2];  /* table+'.'+column+'\0' */

    if (e->op == O_COL) {
	if (strchr (e->e.lit_name, '.'))    /* already GOT table/corr name */
	    return (WALK);
	
	sprintf (newname, "%s.%s", Table_name, e->e.lit_name);
	e->e.lit_name = erealloc (e->e.lit_name, strlen (newname)+1);
	strcpy (e->e.lit_name, newname);
    }
    return (WALK);
}

/* new_del --- make a new delete enode */

enode_t *
new_del (fileloc_t *floc, elist_t *table_list, enode_t *where_clause)
{
    dml_t *d;
    enode_t *e;

    d = (dml_t *) emalloc (sizeof (dml_t));
    d->type = DELETE;
    d->lineno = floc->lineno;
    d->fname = floc->fname;
    free (floc);

    d->table_list = lton (table_list);
    d->column_list = NULL;
    d->where_clause = where_clause;
    d->ins_val = NULL;
    d->asgn_list = NULL;
    d->order_clause = NULL;
    d->sub_sel = NULL;
    d->from_clause = NULL;
    e = (enode_t *) emalloc (sizeof (enode_t));
    e->op = O_DML;
    e->e.dml = d;
    return (e);
}

/* new_upd --- make a new update enode */

enode_t *
new_upd (fileloc_t *floc, elist_t *table_list, elist_t *from_clause,
  elist_t *asgn_list, enode_t *where_clause)
{
    dml_t *d;
    enode_t *e;
    void mod_col_name (elist_t *, elist_t *);

    d = (dml_t *) emalloc (sizeof (dml_t));
    d->type = UPDATE;
    d->lineno = floc->lineno;
    d->fname = floc->fname;
    free (floc);

    mod_col_name (table_list, asgn_list);
    d->table_list = lton (table_list);
    d->column_list = NULL;
    d->where_clause = where_clause;
    d->ins_val = NULL;
    d->asgn_list = lton (asgn_list);
    d->order_clause = NULL;
    d->sub_sel = NULL;
    if (from_clause != NULL)
	d->from_clause = lton (from_clause);
    else
	d->from_clause = NULL;
    e = (enode_t *) emalloc (sizeof (enode_t));
    e->op = O_DML;
    e->e.dml = d;
    return (e);
}

/* new_sel --- make a new select enode */

enode_t *
new_sel (enode_t *sub_sel, elist_t *order_clause)
{
    dml_t *d;
    enode_t *e;

    d = (dml_t *) emalloc (sizeof (dml_t));
    d->type = SELECT;

    find_floc (sub_sel, &d->lineno, &d->fname);

    d->table_list = NULL;
    d->column_list = NULL;
    d->where_clause = NULL;
    d->ins_val = NULL;
    d->asgn_list = NULL;
    if (order_clause)
	d->order_clause = lton (order_clause);
    else
	d->order_clause = NULL;
    d->sub_sel = sub_sel;
    d->from_clause = NULL;
    e = (enode_t *) emalloc (sizeof (enode_t));
    e->op = O_DML;
    e->e.dml = d;
    return (e);
}

/* new_order --- make a new "ORDER BY" list */

elist_t *
new_order (elist_t *order_list, enode_t *name, int dir)
{
    enode_t *order;

    order = (enode_t *) emalloc (sizeof (enode_t));
    order->op = O_ORD;
    order->e.ord.ord_name = name;
    order->e.ord.ord_dir = dir;
    return (new_elist (order_list, order));
}

/* new_subsel --- make a new sub select enode */

enode_t *
new_subsel (fileloc_t *floc, int all_dist_ind, elist_t *result_list,
  enode_t *into, elist_t *from_clause, enode_t *where_clause,
  elist_t *group_clause, enode_t *having_clause)
{
    subsel_t *s;
    enode_t *e;

    s = emalloc (sizeof (subsel_t));
    s->all_dist_ind = all_dist_ind;

    if (result_list)
	s->result_list = lton (result_list);
    else
	s->result_list = NULL;
	
    s->into = into;
	
    if (from_clause)
	s->from_clause = lton (from_clause);
    else
	s->from_clause = NULL;
	
    s->where_clause = where_clause;

    if (group_clause)
	s->group_clause = lton (group_clause);
    else
	s->group_clause = NULL;
	
    s->having_clause = having_clause;
    s->floc = floc;

    e = emalloc (sizeof (enode_t));
    e->op = O_SUBSEL;
    e->e.subsel = s;
    return (e);
}

/* new_alias --- make a new alias */

enode_t *
new_alias (string_t table_name, string_t corr_name)
{
    enode_t *e;

    if (table_name == NULL)
	table_name = strdup ("");
    if (corr_name == NULL)
	corr_name = strdup ("");
    e = emalloc (sizeof (enode_t));
    e->op = O_ALIAS;
    e->e.alias.table_name = table_name;
    e->e.alias.corr_name = corr_name;
    locase (table_name);
    locase (corr_name);
    return (e);
}

/* lton --- make a list into a node */

enode_t *
lton (elist_t *list)
{
    enode_t *e;

    e = (enode_t *) emalloc (sizeof (enode_t));
    e->op = O_LIST;
    e->e.list = list;
    return (e);
}

/* new_elist --- add to a (possibly new) elist */

elist_t *
new_elist (elist_t *elist, enode_t *enode)
{
    if (elist == NULL) {
	elist = (elist_t *) emalloc (sizeof (elist_t));
	elist->size = START_SIZE;
	elist->used = 1;
	elist->elist = (enode_t **) emalloc (sizeof (enode_t *) * elist->size);
	elist->elist[0] = enode;
    }
    else if (elist->used >= elist->size) {
	elist->size += INC_SIZE;
	elist->elist = (enode_t **) erealloc ((void *) elist->elist,
	  sizeof (enode_t *) * elist->size);
	elist->elist[elist->used] = enode;
	elist->used += 1;
    }
    else {
	elist->elist[elist->used] = enode;
	elist->used += 1;
    }
    return (elist);
}

/* new_any --- make a new ANY enode */

enode_t *
new_any (int op)
{
    enode_t *e;

    e = emalloc (sizeof (enode_t));
    e->op = ANY;
    e->e.any_all.cmp = op;
    e->e.any_all.subsel = NULL;
    return (e);
}

/* new_all --- make a new ALL enode */

enode_t *
new_all (int op)
{
    enode_t *e;

    e = emalloc (sizeof (enode_t));
    e->op = ALL;
    e->e.any_all.cmp = op;
    e->e.any_all.subsel = NULL;
    return (e);
}

/* new_like --- return a new LIKE enode */

enode_t *
new_like (enode_t *expr, int like, enode_t *lit, enode_t *esc)
{
    enode_t *e;

    e = emalloc (sizeof (enode_t));
    e->op = LIKE;
    e->e.like.expr = expr;
    e->e.like.negate = !like;
    e->e.like.lit = lit;
    e->e.like.esc = esc;
    return (e);
}

/* new_between --- return a new BETWEEN enode */

enode_t *
new_between (enode_t *expr, int between, enode_t *left, enode_t *right)
{
    enode_t *e;

    e = emalloc (sizeof (enode_t));
    e->op = BETWEEN;
    e->e.between.expr = expr;
    e->e.between.negate = !between;
    e->e.between.left = left;
    e->e.between.right = right;
    return (e);
}

/* new_in --- return a new IN enode */

enode_t *
new_in (enode_t *expr, int in, enode_t *vals)
{
    enode_t *e;

    e = emalloc (sizeof (enode_t));
    e->op = IN;
    e->e.in.expr = expr;
    e->e.in.negate = !in;
    e->e.in.vals = vals;
    return (e);
}

static long xlineno;
static char *xfname;

/* find_floc --- find the fname and lineno for this select */

static void
find_floc (enode_t *e, long *lineno, char **fname)
{
    xfname = NULL;
    (void) walk (e, floc_visit);

    *lineno = xlineno;
    *fname = xfname;
}

/* floc_visit --- find the lowest lineno for each O_SUBSEL */

static int
floc_visit (enode_t *e)
{
    if (e->op == O_SUBSEL) {
	if (xfname == NULL) {
	    xfname = e->e.subsel->floc->fname;
	    xlineno = e->e.subsel->floc->lineno;
	}
	else if (xlineno > e->e.subsel->floc->lineno)
	    xlineno = e->e.subsel->floc->lineno;
	
    }
    return (WALK);
}

/* resize_elist --- realloc each elist */

void
resize_elist ()
{
    int i;
    static int elist_walk (enode_t *);

    for (i = 0; i < DML->used; i++)
	(void) walk (DML->elist[i], elist_walk);
}

/* elist_walk --- realloc one elist node */

static int
elist_walk (enode_t *e)
{
    elist_t *l;

    if (e->op == O_LIST) {
	l = e->e.list;
	if (l->size > l->used) {    /* don't waste memory */
	    l->elist = erealloc (l->elist, l->used * sizeof (enode_t *));
	    l->size = l->used;
	}
    }
    return (WALK);
}
