/* store.h --- structs for storing DMLs for reporting later */
/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

typedef struct {
    char    *fname;
    int     lineno;
} fileloc_t;

typedef struct {
    short   size;       /* enode_t's allocated */
    short   used;       /* enode_t's in use */
    struct  enode   **elist;    /* array of enode_t's */
} elist_t;

# define START_SIZE     2   /* start of elist_t.size */
# define INC_SIZE       4   /* increment of elist_t.size */

typedef struct enode {
    int     op;             /* operation: '+', '-', O_??? see below */
    union {
	string_t    lit_name;       /* literal or column name */
	struct {
	    struct enode *left,     /* operands of op */
			 *right;
	} child;
	struct {
	    string_t func_name;     /* function name */
	    struct enode *arg_list; /* list of enodes to func */
	} func;
	struct {
	    struct  enode   *ord_name;  /* "order by" name */
	    int     ord_dir;        /* direction: ascending/descending */
	} ord;
	struct {
	    string_t    table_name;
	    string_t    corr_name;
	} alias;
	elist_t *list;              /* list of enodes */
	struct Subsel *subsel;      /* sub select */
	struct Dml *dml;            /* SELECT/UPDATE/INSERT/DELETE */
	struct {
	    int     cmp;
	    struct  enode   *subsel;
	} any_all;
	struct {
	    struct  enode   *expr;
	    int     negate;
	    struct  enode   *lit;
	    struct  enode   *esc;
	} like;
	struct {
	    struct  enode   *expr;
	    int     negate;
	    struct  enode   *left;
	    struct  enode   *right;
	} between;
	struct {
	    struct  enode   *expr;
	    int     negate;
	    struct  enode   *vals;
	} in;
    }       e;
} enode_t;

typedef struct Subsel {
    int         all_dist_ind;
    enode_t     *result_list;
    enode_t     *into;
    enode_t     *from_clause;
    enode_t     *where_clause;
    enode_t     *group_clause;
    enode_t     *having_clause;
    fileloc_t   *floc;
} subsel_t;

typedef struct Dml {
    int     type;   /* SELECT, UPDATE, INSERT, DELETE */
    long    lineno;
    char    *fname;
    enode_t *table_list;
    enode_t *column_list;
    enode_t *where_clause;
    enode_t *ins_val;
    enode_t *asgn_list;
    enode_t *order_clause;
    enode_t *sub_sel;
    enode_t *from_clause;
} dml_t;

/* defines for op above */
# define O_LIT      'N'     /* literal e.g. "5", "'FUBAR'", etc */
# define O_COL      'C'     /* column name */
# define O_FUNC     'F'     /* built-in function e.g. "trim", "dow" */
# define O_ORD      'O'     /* order by description */
# define O_ALIAS    'A'     /* tablename & correlation name */
# define O_LIST     'L'     /* enode list */
# define O_SUBSEL   'S'     /* sub select */
# define O_WHERE    'W'     /* where clause */
# define O_DML      'D'     /* DML statement */

extern enode_t *new_enode (int, enode_t *, enode_t *);
extern enode_t *new_lit (enode_t *, string_t, string_t, string_t);
extern enode_t *lit (string_t);
extern enode_t *col (string_t);
extern enode_t *cptr (string_t, string_t);
extern enode_t *bltin (string_t, elist_t *);
extern elist_t *new_elist (elist_t *, enode_t *);
extern enode_t *new_ins (fileloc_t *, elist_t *, elist_t *, enode_t *);
extern enode_t *new_del (fileloc_t *, elist_t *, enode_t *);
extern enode_t *new_upd (fileloc_t *, elist_t *, elist_t *, elist_t *, enode_t *);
extern elist_t *new_order (elist_t *, enode_t *, int);
extern enode_t *new_subsel (fileloc_t *, int, elist_t *, enode_t *,
  elist_t *, enode_t *, elist_t *, enode_t *);
extern enode_t *new_alias (string_t, string_t);
extern enode_t *new_sel (enode_t *, elist_t *);
extern enode_t *lton (elist_t *);
extern enode_t *new_any (int);
extern enode_t *new_all (int);
extern enode_t *new_like (enode_t *, int, enode_t *, enode_t *);
extern enode_t *new_between (enode_t *, int, enode_t *, enode_t *);
extern enode_t *new_in (enode_t *, int, enode_t *);
extern void     store (enode_t *);
extern void    *emalloc (int), *erealloc (void *, int);
extern elist_t *DML;
extern int  walk (enode_t *, int (*) (enode_t *));
extern void treeprint (enode_t *, FILE *);
