/* util.c --- misc helpful functions */
/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

# include <stdio.h>
# include <ctype.h>
# include <stdlib.h>
# include "ferret.h"
# include "store.h"
# ifdef VMS
#	include "y_tab.h"
# else
#	include "y.tab.h"
# endif

extern char *progname;

int isnumber (char *);

/* ztop --- change zeros to periods */

char *
ztop (long n)
{
    static char buf[20][20];
    static int i = 0;
    int cur;

    if (!n)
	return (".");
    else
	sprintf (buf[i], "%ld", n);

    cur = i;
    if (++i >= 20)
	i = 0;
    return (buf[cur]);
}

/* putnc --- put N c's to fp */

void
putnc (int n, int c, FILE *fp)
{
    while (n--)
	putc (c, fp);
}

/* emalloc --- wrapper for malloc, die on error */

void *
emalloc (int size)
{
    void *ret;

    ret = malloc ((unsigned) size);
    if (ret == NULL) {
	fprintf (stderr, "%s: out of mem!\n", progname);
	exit (1);
    }
    return (ret);
}

/* erealloc --- wrapper for realloc, die on error */

void *
erealloc (void *ptr, int size)
{
    void *ret;

    ret = realloc (ptr, (unsigned) size);
    if (ret == NULL) {
	fprintf (stderr, "%s: out of mem!\n", progname);
	exit (1);
    }
    return (ret);
}

/* locase --- convert string to lowercase */

void
locase (string_t s)
{
    if (!s)
	return;	
    while (*s) {
	if (isupper (*s))
	    *s = tolower (*s);
	++s;
    }
}

/* walk --- recursively walk the enodes, calling func with each */

int
walk (enode_t *e, int (*func) (enode_t *))
{
    if (e == NULL)
	return (WALK);
	
    if ((*func) (e) == STOP)
	return (STOP);
	
    switch (e->op) {
    case '+':
    case '-':
    case '*':
    case '/':
    case GT:
    case GE:
    case LT:
    case LE:
    case EQ:
    case NE:
    case AND:
    case OR:
    case ASGN:
    case UNION:
    case UNION_ALL:
	if (walk (e->e.child.left, func) == STOP)
	    return (STOP);
	if (walk (e->e.child.right, func) == STOP)
	    return (STOP);
	break;
    case NOT:
    case UMINUS:
    case EXISTS:
    case O_WHERE:
	if (walk (e->e.child.left, func) == STOP)
	    return (STOP);
	break;
    case O_LIST:
	{
	    int i;
	    elist_t *l;
	
	    l = e->e.list;
	    if (l == NULL)
		break;
	    for (i = 0; i < l->used; i++)
		if (walk (l->elist[i], func) == STOP)
		    return (STOP);
	}
	break;
    case O_SUBSEL:
	{
	    subsel_t *s;
	    s = e->e.subsel;
	    if (walk (s->result_list, func) == STOP)
		return (STOP);
	    if (s->from_clause) {
		if (walk (s->from_clause, func) == STOP)
		    return (STOP);
		
	    }
	    if (s->into) {
		if (walk (s->into, func) == STOP)
		    return (STOP);
	    }
	    if (s->where_clause) {
		if (walk (s->where_clause, func) == STOP)
		    return (STOP);
	    }
	    if (s->group_clause) {
		if (walk (s->group_clause, func) == STOP)
		    return (STOP);
	    }
	    if (s->having_clause) {
		if (walk (s->having_clause, func) == STOP)
		    return (STOP);
	    }
	}
	break;
    case O_DML:
	{
	    dml_t *d;
	    d = e->e.dml;
	    if (walk (d->table_list, func) == STOP)
		return (STOP);
	    if (walk (d->column_list, func) == STOP)
		return (STOP);
	    if (walk (d->where_clause, func) == STOP)
		return (STOP);
	    if (walk (d->ins_val, func) == STOP)
		return (STOP);
	    if (walk (d->asgn_list, func) == STOP)
		return (STOP);
	    if (walk (d->order_clause, func) == STOP)
		return (STOP);
	    if (walk (d->sub_sel, func) == STOP)
		return (STOP);
	    if (walk (d->from_clause, func) == STOP)
		return (STOP);
	}
	break;
    case O_LIT:
    case O_COL:
	break;
    case O_ORD:
	if (walk (e->e.ord.ord_name, func) == STOP)
	    return (STOP);
	break;
    case O_ALIAS:
	break;
    case O_FUNC:
	if (walk (e->e.func.arg_list, func) == STOP)
	    return (STOP);
	break;
    case INTO:
	if (walk (e->e.child.left, func) == STOP)
	    return (STOP);
	break;
    case LIKE:
	if (walk (e->e.like.expr, func) == STOP)
	    return (STOP);
	break;
    case BETWEEN:
	if (walk (e->e.between.expr, func) == STOP)
	    return (STOP);
	if (walk (e->e.between.left, func) == STOP)
	    return (STOP);
	if (walk (e->e.between.right, func) == STOP)
	    return (STOP);
	break;
    case IN:
	if (walk (e->e.in.expr, func) == STOP)
	    return (STOP);
	if (walk (e->e.in.vals, func) == STOP)
	    return (STOP);
	break;
    case ANY:
    case ALL:
	if (walk (e->e.any_all.subsel, func) == STOP)
	    return (STOP);
	break;
    case SQL_NULL:
    case NOT_SQL_NULL:
	if (walk (e->e.child.left, func) == STOP)
	    return (STOP);
	break;
    default:
	fprintf (stderr, "%s: unknown op value: %d",
	  progname, e->op);
	if (isprint (e->op))
	    fprintf (stderr, " (%c)", e->op);
	putc ('\n', stderr);
	abort ();
	break;
    }
    return (WALK);
}

/* print_dml --- print each dml */

void
print_dml ()
{
    int i;

    printf ("%d DML%s\n", DML->used, (DML->used > 1) ? "s" : "");
    for (i = 0; i < DML->used; i++) {
	printf ("DML #%d\n", i+1);
	treeprint (DML->elist[i], stdout);
	printf ("\n\n");
    }
}

/* treeprint --- recursively print the enodes */

void
treeprint (enode_t *e, FILE *fp)
{
    if (e == NULL)
	return;
	
    switch (e->op) {
    case '+':
    case '-':
    case '*':
    case '/':
	treeprint (e->e.child.left, fp);
	fprintf (fp, " %c ", e->op);
	treeprint (e->e.child.right, fp);
	break;
    case GT:
	treeprint (e->e.child.left, fp);
	fprintf (fp, " > ");
	treeprint (e->e.child.right, fp);
	break;
    case GE:
	treeprint (e->e.child.left, fp);
	fprintf (fp, " >= ");
	treeprint (e->e.child.right, fp);
	break;
    case LT:
	treeprint (e->e.child.left, fp);
	fprintf (fp, " < ");
	treeprint (e->e.child.right, fp);
	break;
    case LE:
	treeprint (e->e.child.left, fp);
	fprintf (fp, " <= ");
	treeprint (e->e.child.right, fp);
	break;
    case EQ:
	treeprint (e->e.child.left, fp);
	fprintf (fp, " = ");
	treeprint (e->e.child.right, fp);
	break;
    case NE:
	treeprint (e->e.child.left, fp);
	fprintf (fp, " != ");
	treeprint (e->e.child.right, fp);
	break;
    case AND:
	treeprint (e->e.child.left, fp);
	fprintf (fp, "\n     AND ");
	treeprint (e->e.child.right, fp);
	break;
    case OR:
	treeprint (e->e.child.left, fp);
	fprintf (fp, "\n      OR ");
	treeprint (e->e.child.right, fp);
	break;
    case ASGN:
	treeprint (e->e.child.left, fp);
	fprintf (fp, " = ");
	treeprint (e->e.child.right, fp);
	break;
    case UNION:
	treeprint (e->e.child.left, fp);
	fprintf (fp, "\n\nUNION\n\n");
	treeprint (e->e.child.right, fp);
	break;
    case UNION_ALL:
	treeprint (e->e.child.left, fp);
	fprintf (fp, "\n\nUNION ALL\n\n");
	treeprint (e->e.child.right, fp);
	break;
    case NOT:
	fprintf (fp, " NOT ");
	treeprint (e->e.child.left, fp);
	break;
    case UMINUS:
	fprintf (fp, " -");
	treeprint (e->e.child.left, fp);
	break;
    case O_WHERE:
	fprintf (fp, "\n   WHERE ");
	treeprint (e->e.child.left, fp);
	break;
    case O_LIST:
	{
	    int i;
	    elist_t *l;
	
	    l = e->e.list;
	    if (l == NULL)
		break;
	    for (i = 0; i < l->used; i++) {
		treeprint (l->elist[i], fp);
		if (i+1 < l->used)
		    fprintf (fp, ", ");
	    }
	}
	break;
    case O_SUBSEL:
	{
	    subsel_t *s;
	    s = e->e.subsel;
	    fprintf (fp, "  SELECT ");
	    treeprint (s->result_list, fp);
	    if (s->into) {
		treeprint (s->into, fp);
	    }
	    if (s->from_clause) {
		fprintf (fp, "\n    FROM ");
		treeprint (s->from_clause, fp);
	    }
	    if (s->where_clause) {
		treeprint (s->where_clause, fp);
	    }
	    if (s->group_clause) {
		fprintf (fp, "\nGROUP BY ");
		treeprint (s->group_clause, fp);
	    }
	    if (s->having_clause) {
		fprintf (fp, "\n  HAVING ");
		treeprint (s->having_clause, fp);
	    }
	}
	break;
    case O_DML:
	{
	    dml_t *d;
	    d = e->e.dml;
	    switch (d->type) {
	    case DELETE:
		fprintf (fp, "DELETE FROM ");
		break;
	    case INSERT:
		fprintf (fp, "INSERT INTO ");
		break;
	    case UPDATE:
		fprintf (fp, "UPDATE ");
		break;
	    }
	    if (d->table_list) {
		treeprint (d->table_list, fp);
		putc ('\n', fp);
	    }
	    if (d->from_clause) {
		fprintf (fp, "\n    FROM ");
		treeprint (d->from_clause, fp);
		putc ('\n', fp);
	    }
	    if (d->column_list) {
		fprintf (fp, " (");
		treeprint (d->column_list, fp);
		fprintf (fp, ")\n");
	    }
	    if (d->asgn_list) {
		fprintf (fp, "SET ");
		treeprint (d->asgn_list, fp);
		putc ('\n', fp);
	    }
	    if (d->where_clause) {
		treeprint (d->where_clause, fp);
		putc ('\n', fp);
	    }
	    if (d->ins_val) {
		treeprint (d->ins_val, fp);
		putc ('\n', fp);
	    }
	    if (d->sub_sel) {
		treeprint (d->sub_sel, fp);
		putc ('\n', fp);
	    }
	    if (d->order_clause) {
		fprintf (fp, "ORDER BY ");
		treeprint (d->order_clause, fp);
		putc ('\n', fp);
	    }
	}
	break;
    case O_LIT:
	if (*e->e.lit_name == ':' || isnumber (e->e.lit_name))
	    fprintf (fp, "%s", e->e.lit_name);
	else
	    fprintf (fp, "'%s'", e->e.lit_name);
	break;
    case O_COL:
	fprintf (fp, "%s", e->e.lit_name);
	break;
    case O_ORD:
	treeprint (e->e.ord.ord_name, fp);
	fprintf (fp, " %s", (e->e.ord.ord_dir == ASC) ? "ASC" : "DESC");
	break;
    case O_ALIAS:
	fprintf (fp, "%s", e->e.alias.table_name);
	if (e->e.alias.corr_name)
	    fprintf (fp, " %s", e->e.alias.corr_name);
	break;
    case O_FUNC:
	fprintf (fp, "%s (", e->e.func.func_name);
	treeprint (e->e.func.arg_list, fp);
	fprintf (fp, ")");
	break;
    case INTO:
	fprintf (fp, "\nINTO ");
	treeprint (e->e.child.left, fp);
	break;
    case LIKE:
	treeprint (e->e.like.expr, fp);
	if (e->e.like.negate)
	    fprintf (fp, " NOT");
	fprintf (fp, " LIKE ");
	treeprint (e->e.like.lit, fp);
	if (e->e.like.esc) {
	    fprintf (fp, " ESCAPE ");
	    treeprint (e->e.like.esc, fp);
	}
	break;
    case EXISTS:
	fprintf (fp, "EXISTS ( ");
	treeprint (e->e.child.left, fp);
	fprintf (fp, ") ");
	break;
    case BETWEEN:
	treeprint (e->e.between.expr, fp);
	if (e->e.between.negate)
	    fprintf (fp, " NOT");
	fprintf (fp, " BETWEEN ");
	treeprint (e->e.between.left, fp);
	fprintf (fp, " AND ");
	treeprint (e->e.between.right, fp);
	break;
    case IN:
	treeprint (e->e.in.expr, fp);
	if (e->e.in.negate)
	    fprintf (fp, " NOT");
	fprintf (fp, " IN (");
	treeprint (e->e.in.vals, fp);
	fprintf (fp, ") ");
	break;
    case ANY:
    case ALL:
	switch (e->e.any_all.cmp) {
	case EQ:
	    fprintf (fp, " =");
	    break;
	case LT:
	    fprintf (fp, " <");
	    break;
	case LE:
	    fprintf (fp, " <=");
	    break;
	case GT:
	    fprintf (fp, " >");
	    break;
	case GE:
	    fprintf (fp, " >=");
	    break;
	case NE:
	    fprintf (fp, " <>");
	}
	if (e->op == ANY)
	    fprintf (fp, "ANY (");
	else
	    fprintf (fp, "ALL (");
	treeprint (e->e.any_all.subsel, fp);
	fprintf (fp, ") ");
	break;
    case SQL_NULL:
    case NOT_SQL_NULL:
	treeprint (e->e.child.left, fp);
	fprintf (fp, " IS");
	if (e->op == NOT_SQL_NULL)
	    fprintf (fp, " NOT");
	fprintf (fp, " NULL ");
	break;
    default:
	fprintf (stderr, "%s: unknown op value: %d",
	  progname, e->op);
	if (isprint (e->op))
	    fprintf (stderr, " (%c)", e->op);
	putc ('\n', stderr);
	abort ();
	break;
    }
}

/* isnumber --- does this look like a number? */

int
isnumber (register char *word)
{
    register int ndot = 0, nexp = 0, ndigit = 0;

    if (!isdigit (*word) && *word != '.')
	return (0);
	
    while (*word) {
	if (*word == '.')
	    ++ndot;
	else if (*word == 'E' || *word == 'e')
	    ++nexp;
	else if (isdigit (*word))
	    ++ndigit;
	else
	    return (0);
	++word;
    }

    return (ndot <= 1 && nexp <= 1 && ndigit >= 1);
}


/* strdup --- malloc space and copy s into it */

char *
strdup (char *str)
{
	char *new;

	new = emalloc (strlen (str) + 1);
	strcpy (new, str);
	return (new);
}
