/* warnsum.c --- wauning summary reports */
/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <errno.h>
# include "ferret.h"
# include "store.h"
# ifdef VMS
#	include "y_tab.h"
# else
#	include "y.tab.h"
# endif

# define YES        1
# define NO         0

static FILE *fp;

static char *Filename;
static int Lineno;

extern char *progname;

static void col_cmp (void);
static int join_look (enode_t *);
static int func_look (enode_t *);
static void func_where (void);
static void null_from (void);

/* warnsum --- produce the warning summary report */

void
warnsum (string_t outbase, string_t ext)
{
    char fname[256];
    extern int Pagelength;

    if (!DML)
	return;

    sprintf (fname, "%s%c%s", outbase, EXTSEP, ext);
    if ((fp = fopen (fname, "w")) == NULL) {
	fprintf (stderr, "%s: can't open %s for writing (%s)\n",
	  progname, fname, strerror (errno));
	return;
    }

    pg_open (fp, Pagelength, 80, 5, 5);
    pg_head (1, "|WARNING SUMMARY|%D", '|');
    pg_head (2, "|%T", '|');

    col_cmp ();     /* look for JOINed disimilar column names */

    func_where ();  /* look for FUNCtions in WHERE clauses */

    null_from ();   /* look for null FROM clauses in subSELECTs */

    pg_close ();

    fclose (fp);
}

/* col_cmp --- look for join columns with disimilar names */

static void
col_cmp ()
{
    int i;
    static int join_look (enode_t *);

    for (i = 0; i < DML->used; i++) {
	Filename = DML->elist[i]->e.dml->fname;
	Lineno = DML->elist[i]->e.dml->lineno;
	(void) walk (DML->elist[i], join_look);
    }
}

/* join_look --- look for O_COL  EQ  O_COL nodes */

static int
join_look (enode_t *e)
{
    enode_t *l, *r;
    char *l_name, *r_name;

    if (e->op == EQ) {
	if ((l = e->e.child.left) == NULL)
	    return (WALK);
	if ((r = e->e.child.right) == NULL)
	    return (WALK);
	if (l->op == O_COL && r->op == O_COL) {
	    if (l_name = strchr (l->e.lit_name, '.'))
		++l_name;
	    else
		l_name = l->e.lit_name;
	    if (r_name = strchr (r->e.lit_name, '.'))
		++r_name;
	    else
		r_name = r->e.lit_name;
	    if (strcmp (l_name, r_name) != 0) {
		pg_printf (
		  "%s @ %ld: JOIN on disimilar column names (\"%s\", \"%s\")\n",
		  Filename, Lineno, l_name, r_name);
	    }
	}
    }
    return (WALK);
}

/* func_where --- look for functions in the WHERE clauses */

static void
func_where ()
{
    int i;
    static int func_look (enode_t *);

    for (i = 0; i < DML->used; i++) {
	Filename = DML->elist[i]->e.dml->fname;
	Lineno = DML->elist[i]->e.dml->lineno;
	(void) walk (DML->elist[i], func_look);
    }
}

/* func_look --- look at one node for functions */

static int
func_look (enode_t *e)
{
    enode_t *l, *r;
    static int func_walk (enode_t *);

    l = e->e.child.left;
    r = e->e.child.right;
	
    switch (e->op) {
    case EQ:
    case NE:
    case GT:
    case GE:
    case LT:
    case LE:
	(void) func_walk (l);
	(void) func_walk (r);
	break;
    }
    return (WALK);
}

/* func_walk --- walk certain of these enodes, noting funcs */

static int
func_walk (enode_t *e)
{
    switch (e->op) {
	case O_FUNC:
	    if (has_col (e))
		pg_printf ("%s @ %d: function \"%s\" in WHERE clause.\n",
		  Filename, Lineno, e->e.func.func_name);
	    /* walk the funcs's arg list */
	    walk (e->e.func.arg_list, func_walk);
	    break;
	case '+':
	case '-':
	case '*':
	case '/':
	case UMINUS:
	case EXPON:
	    (void) walk (e->e.child.left, func_walk);
	    (void) walk (e->e.child.right, func_walk);
	    break;
    }
    return (WALK);
}

/* has_col --- does this function's arg list have a column? */

static int
has_col (enode_t *e)
{
    int i;
    elist_t *arg;

    arg = e->e.func.arg_list->e.list;
    if (arg == NULL)
	return (0);

    for (i = 0; i < arg->used; i++)
	if (arg->elist[i]->op == O_COL)
	    return (1);

    return (0);
}

/* null_from --- look for NULL from clauses in subselects */

static void
null_from ()
{
    register int i;
    static int from_look (enode_t *);

    for (i = 0; i < DML->used; i++) {
	Filename = DML->elist[i]->e.dml->fname;
	Lineno = DML->elist[i]->e.dml->lineno;
	(void) walk (DML->elist[i], from_look);
    }
}

/* from_look --- look at each node for null FROM clauses */

static int
from_look (enode_t *e)
{
    if (e->op == O_SUBSEL && e->e.subsel->from_clause == NULL)
	pg_printf ("%s @ %d: SELECT without FROM clause.\n", Filename, 
	  Lineno);

    return (WALK);
}
