/* wcs.c --- Where Clause Summary report */
/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

/*
** side effect: calls opt_store on each table/column it finds, so the optimizer
** won't have to do this work again
*/

# include <stdio.h>
# include <string.h>
# include <stdlib.h>
# include <errno.h>
# include "ferret.h"
# include "store.h"
# include "page.h"
# ifdef VMS
#	include "y_tab.h"
# else
#	include "y.tab.h"
# endif

# define WU_EQ      1
# define WU_NE      2
# define WU_RANGE   3
# define WU_SET     4
# define WU_EXPR    5
# define WU_QUAL    6
# define WU_JOIN    7

/* type for Where Clause Summary Report */
typedef struct {
    char    *name;          /* as stored in an enode */
    char    *table_name;
    char    *column_name;
    long    neq;            /* # EQ (col = XX) */
    long    nne;            /* # NE (col != XX) */
    long    nrange;         /* # range (col <= XX or att like .. */
    long    nset;           /* # sets (col in (XX, YY, ZZ)) */
    long    nexpr;          /* # exprs (col + XX) */
    long    nqual;          /* # quals (qualification (col = :var)) */
    long    njoin;          /* # joins (col1 = col2) */
} where_use_t;

static where_use_t **Where_use;     /* array of -> where_use_ts */
static int Nwhere_use = 0;          /* how many where_use_ts */
static int MAXwhere_use = 0;        /* how many allocated */

static int wu_cmp (where_use_t **, where_use_t **);
static where_use_t * wu_find (char *);
static void wu_inc (char *, int);
static int wu_store (enode_t *);

extern void opt_store (char *, char *);

extern char *progname;

void
where_use (string_t outbase, string_t ext)
{
    int i;
    where_use_t *w;
    char fname[256];
    FILE *fp;
    extern int Pagelength;
    int wu_cmp (where_use_t **, where_use_t **);

    if (!DML)
	return;

    /* run through the where clauses, and store tables and columns */
    for (i = 0; i < DML->used; i++)
	(void) walk (DML->elist[i], wu_store);

    qsort (Where_use, Nwhere_use, sizeof (where_use_t *), wu_cmp);

    sprintf (fname, "%s%c%s", outbase, EXTSEP, ext);
    if ((fp = fopen (fname, "w")) == NULL) {
	fprintf (stderr, "%s: can't open %s for writing (%s)\n",
	  progname, fname, strerror (errno));
	return;
    }

    if (Nwhere_use > 0) {
	char buf[128];
	pg_open (fp, Pagelength, 78, 5, 5);
	pg_head (1, "|WHERE CLAUSE SUMMARY|%D", '|');
	pg_head (2, "|%T", '|');
	sprintf (buf, "%-20s %-20s %4s %4s %4s %4s %4s %4s %4s|",
	  "Table",
	  "Column",
	  " =",
	  "!=",
	  "m-n",
	  "(a,b)",
	  "expr",
	  "qual",
	  "join"
	  );
	pg_head (3, buf, '|');
	pg_head (4,
	  "-----------------------------------------------------------------------------|",
	  '|');
    }

    for (i = 0; i < Nwhere_use; i++) {
	w = Where_use[i];
	pg_printf ("%-20.20s %-20.20s %4s %4s %4s %4s %4s %4s %4s\n",
	  w->table_name,
	  w->column_name,
	  ztop (w->neq),
	  ztop (w->nne),
	  ztop (w->nrange),
	  ztop (w->nset),
	  ztop (w->nexpr),
	  ztop (w->nqual),
	  ztop (w->njoin)
	  );
	opt_store (w->table_name, w->column_name);
	free (w->name);
	free (w->table_name);
	free (w->column_name);
	free (w);
    }
    free (Where_use);
    Where_use = 0;
    Nwhere_use = 0;
    MAXwhere_use = 0;
    pg_close ();
    fclose (fp);
}

/* wu_store --- for certain operators, store 'em */

static int
wu_store (enode_t *e)
{
    enode_t *l, *r;

    l = e->e.child.left;
    r = e->e.child.right;
    switch (e->op) {
    case EQ:
	if (l->op == O_COL) {
	    if (r->op == O_COL)
		wu_inc (l->e.lit_name, WU_JOIN);
	    else
		wu_inc (l->e.lit_name, WU_EQ);
	}
	if (r->op == O_COL) {
	    if (l->op == O_COL)
		wu_inc (r->e.lit_name, WU_JOIN);
	    else
		wu_inc (r->e.lit_name, WU_EQ);
	}
	break;
    case NE:
	if (l->op == O_COL)
	    wu_inc (l->e.lit_name, WU_NE);
	if (r->op == O_COL)
	    wu_inc (r->e.lit_name, WU_NE);
	break;
    case GT:
    case GE:
    case LT:
    case LE:
	if (l->op == O_COL)
	    wu_inc (l->e.lit_name, WU_RANGE);
	else if (r->op == O_COL)
	    wu_inc (r->e.lit_name, WU_RANGE);
	break;
    case LIKE:
	if (e->e.like.expr->op == O_COL)
	    wu_inc (e->e.like.expr->e.lit_name, WU_RANGE);
	break;
    case BETWEEN:
	if (e->e.between.left->op == O_COL)
	    wu_inc (e->e.between.left->e.lit_name, WU_RANGE);
	if (e->e.between.right->op == O_COL)
	    wu_inc (e->e.between.right->e.lit_name, WU_RANGE);
	break;
    case IN:
	if (e->e.in.expr->op == O_COL)
	    wu_inc (e->e.in.expr->e.lit_name, WU_SET);
	break;
    case '+':
    case '-':
    case '*':
    case '/':
    case EXPON:
	if (l->op == O_COL)
	    wu_inc (l->e.lit_name, WU_EXPR);
	if (r->op == O_COL)
	    wu_inc (r->e.lit_name, WU_EXPR);
	break;
    case O_FUNC:
	if (strcmp (e->e.func.func_name, "qualification") == 0) {
	    int i;
	    elist_t *l;
	    enode_t *a;
	    l = e->e.func.arg_list->e.list;
	    for (i = 0; i < l->used; ++i) {
		a = l->elist[i];
		if (a->op == EQ) {
		    if (a->e.child.left->op == O_COL)
			wu_inc (a->e.child.left->e.lit_name, WU_QUAL);
		    if (a->e.child.right->op == O_COL)
			wu_inc (a->e.child.right->e.lit_name, WU_QUAL);
		}
		if (a->op == ASGN) {
		    if (a->e.child.left->op == O_COL)
			wu_inc (a->e.child.left->e.lit_name, WU_QUAL);
		}
	    }
	}
	break;
    }
    return (WALK);
}

/* wu_inc --- increment the appropriate element for this table/col */

static void
wu_inc (char *name, int element)
{
    where_use_t *w;

    w = wu_find (name);
    switch (element) {
    case WU_EQ:
	w->neq += 1;
	break;
    case WU_NE:
	w->nne += 1;
	break;
    case WU_RANGE:
	w->nrange += 1;
	break;
    case WU_SET:
	w->nset += 1;
	break;
    case WU_EXPR:
	w->nexpr += 1;
	break;
    case WU_QUAL:
	w->nqual += 1;
	break;
    case WU_JOIN:
	w->njoin += 1;
	break;
    }
}

/* wu_find --- return a (possibly new) where_use_t */

static where_use_t *
wu_find (char *name)
{
    int i;
    char *p;
	
    for (i = 0; i < Nwhere_use; i++)
	if (strcmp (name, Where_use[i]->name) == 0)
	    return (Where_use[i]);

    if (Nwhere_use >= MAXwhere_use) {
	MAXwhere_use += 10;
	if (Where_use == NULL)
	    Where_use = emalloc (MAXwhere_use * sizeof (Where_use[0]));
	else
	    Where_use = erealloc (Where_use,
	      MAXwhere_use * sizeof (Where_use[0]));
    }
	
    i = Nwhere_use++;
    Where_use[i] = emalloc (sizeof (where_use_t));
    memset (Where_use[i], 0, sizeof (where_use_t));
    Where_use[i]->name = strdup (name);
    if ((p = strchr (name, '.'))) {
	*p = 0;
	Where_use[i]->table_name = strdup (name);
	Where_use[i]->column_name = strdup (p+1);
	*p = '.';
    }
    else {
	Where_use[i]->column_name = strdup (name);
	Where_use[i]->table_name = strdup ("*** AMBIGUOUS ***");
    }
    return (Where_use[i]);
}

/* wu_cmp --- compare two elements for qsort */

static int
wu_cmp (where_use_t **a, where_use_t **b)
{
    int atot, btot;

    atot = (*a)->neq + (*a)->nne + (*a)->nrange + (*a)->nset +
      (*a)->nexpr + (*a)->nqual + (*a)->njoin;
    btot = (*b)->neq + (*b)->nne + (*b)->nrange + (*b)->nset +
      (*b)->nexpr + (*b)->nqual + (*b)->njoin;
    if (atot != btot)
	return (btot - atot);
    else
	return (strcmp ((*a)->name, (*b)->name));
}
