/* xref.c --- table/column cross reference */
/*
		Copyright 1994 In Touch Ministries
		Written by Daniel S. Cox (Danny@itm.org)


			   NO WARRANTY


  BECAUSE THIS PROGRAM IS LICENSED FREE OF CHARGE, WE PROVIDE ABSOLUTELY
NO WARRANTY, TO THE EXTENT PERMITTED BY APPLICABLE STATE LAW.  EXCEPT
WHEN OTHERWISE STATED IN WRITING, IN TOUCH MINISTRIES AND/OR OTHER
PARTIES PROVIDE THIS PROGRAM "AS IS" WITHOUT WARRANTY OF ANY KIND,
EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE
ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.
SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY
SERVICING, REPAIR OR CORRECTION.

 IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW WILL DANIEL S. COX,
IN TOUCH MINISTRIES, AND/OR ANY OTHER PARTY WHO MAY MODIFY AND
REDISTRIBUTE THIS PROGRAM AS PERMITTED BELOW, BE LIABLE TO YOU FOR
DAMAGES, INCLUDING ANY LOST PROFITS, LOST MONIES, OR OTHER SPECIAL,
INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE OR INABILITY
TO USE (INCLUDING BUT NOT LIMITED TO LOSS OF DATA OR DATA BEING RENDERED
INACCURATE OR LOSSES SUSTAINED BY THIRD PARTIES OR A FAILURE OF THE
PROGRAM TO OPERATE WITH ANY OTHER PROGRAMS) THIS PROGRAM, EVEN IF YOU
HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES, OR FOR ANY CLAIM
BY ANY OTHER PARTY.


		GENERAL PUBLIC LICENSE TO COPY


  1. You may copy and distribute verbatim copies of this source file
as you receive it, in any medium, provided that you conspicuously and
appropriately publish on each copy a valid copyright notice "Copyright
1993 In Touch Ministries"; and include following the
copyright notice a verbatim copy of the above disclaimer of warranty
and of this License.  You may charge a distribution fee for the
physical act of transferring a copy.

  2. You may modify your copy or copies of this source file or
any portion of it, and copy and distribute such modifications under
the terms of Paragraph 1 above, provided that you also do the following:

    a) cause the modified files to carry prominent notices stating
    that you changed the files and the date of any change; and

    b) cause the whole of any work that you distribute or publish,
    that in whole or in part contains or is a derivative of this
    program or any part thereof, to be licensed at no charge to all
    third parties on terms identical to those contained in this
    License Agreement (except that you may choose to grant more extensive
    warranty protection to some or all third parties, at your option).

    c) You may charge a distribution fee for the physical act of
    transferring a copy, and you may at your option offer warranty
    protection in exchange for a fee.

Mere aggregation of another unrelated program with this program (or its
derivative) on a volume of a storage or distribution medium does not bring
the other program under the scope of these terms.

  3. You may copy and distribute this program or any portion of it in
compiled, executable or object code form under the terms of Paragraphs
1 and 2 above provided that you do the following:

    a) accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    b) accompany it with a written offer, valid for at least three
    years, to give any third party free (except for a nominal
    shipping charge) a complete machine-readable copy of the
    corresponding source code, to be distributed under the terms of
    Paragraphs 1 and 2 above; or,

    c) accompany it with the information you received as to where the
    corresponding source code may be obtained.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form alone.)

For an executable file, complete source code means all the source code for
all modules it contains; but, as a special exception, it need not include
source code for modules which are standard libraries that accompany the
operating system on which the executable file runs.

  4. You may not copy, sublicense, distribute or transfer this program
except as expressly provided under this License Agreement.  Any attempt
otherwise to copy, sublicense, distribute or transfer this program is void and
your rights to use the program under this License agreement shall be
automatically terminated.  However, parties who have received computer
software programs from you with this License Agreement will not have
their licenses terminated so long as such parties remain in full compliance.

In other words, you are welcome to use, share and improve this program.
You are forbidden to forbid anyone else to use, share and improve
what you give them.   Help stamp out software-hoarding!
*/

# include <stdio.h>
# include <stdlib.h>
# include <string.h>
# include <errno.h>
# include "ferret.h"
# include "store.h"
# ifdef VMS
#	include "y_tab.h"
# else
#	include "y.tab.h"
# endif

/* type for table/column cross reference */

typedef struct {
    char    *name;          /* like "table.column" */
    char    *table_name;    /* like "table" */
    char    *column_name;   /* like "column" */
    char    *file_name;     /* source file name */
    long    lineno;         /* source line number */
    int     type;           /* INSERT, DELETE, etc */
} cross_ref_t;

static cross_ref_t **Cross_ref;     /* array of -> cross_ref_ts */
static int Ncross_ref = 0;          /* how many cross_ref_ts */
static int MAXcross_ref = 0;        /* how many allocated */

static char *file_name;
static long lineno;
static int type;

static char *dotype (int);
static int cr_cmp (cross_ref_t **, cross_ref_t **);
static cross_ref_t *cr_new (char *);
static void cr_add (char *);
static int cr_store (enode_t *);
static int cr_walk (enode_t *);

static FILE *fp;

extern char *progname;

void
cross_ref (string_t outbase, string_t ext)
{
    int i;
    cross_ref_t *c;
    char fname[256];
    extern int Pagelength;

    if (!DML)
	return;

    sprintf (fname, "%s%c%s", outbase, EXTSEP, ext);
    if ((fp = fopen (fname, "w")) == NULL) {
	fprintf (stderr, "%s: can't open %s for writing (%s)\n",
	  progname, fname, strerror (errno));
	return;
    }

    /* run through the where clauses, and store tables and columns */
    for (i = 0; i < DML->used; i++) {
	(void) walk (DML->elist[i], cr_walk);
    }
/**********************
    qsort (Cross_ref, Ncross_ref, sizeof (cross_ref_t *), cr_cmp);
********************/

# ifdef COMMENT
    if (Ncross_ref > 0) {
	char line[128];
	
	pg_open (fp, Pagelength, 85, 5, 5);
	pg_head (1, "|TABLE_COLUMN CROSS REFERENCE|%D", '|');
	pg_head (2, "|%T", '|');
	sprintf (line, "%-20.20s %-20.20s %-30s %4s %s|",
	  "Table",
	  "Column",
	  "File",
	  "Line",
	  "Type"
	  );
	pg_head (3, line, '|');
	pg_head (4,
	  "-----------------------------------------------------------------------------------|",
	  '|');
    }

    for (i = 0; i < Ncross_ref; i++) {
	c = Cross_ref[i];
if (c->table_name <= 200)
    fprintf (stderr, "table_name bad, file %s line %d\n", c->file_name,
      c->lineno);
if (c->column_name <= 200)
    fprintf (stderr, "column_name bad, file %s line %d\n", c->file_name,
      c->lineno);
	if (c->table_name <= 200)
	    c->table_name = strdup ("");
	if (c->column_name <= 200)
	    c->column_name = strdup ("");
	if (c->file_name <= 200)
	    c->file_name = "";
	pg_printf ("%-20.20s %-20.20s %-30s %4ld %s\n",
	  c->table_name,
	  c->column_name,
	  c->file_name,
	  c->lineno,
	  dotype (c->type)
	  );
	if (c->table_name)
	    free (c->table_name);
	if (c->column_name)
	    free (c->column_name);
	free (c);
    }

# endif /* COMMENT */

    free (Cross_ref);
    Cross_ref = NULL;

    pg_close ();

    fclose (fp);
}

/* cr_walk --- called for each node of each DML */

static int
cr_walk (enode_t *e)
{
    int cr_store (enode_t *);

    if (e->op == O_DML) {
	file_name = e->e.dml->fname;
	lineno = e->e.dml->lineno;
	type = e->e.dml->type;
	(void) walk (e->e.dml->column_list, cr_store);
	(void) walk (e->e.dml->asgn_list, cr_store);
	(void) walk (e->e.dml->where_clause, cr_store);
    }
    else if (e->op == O_SUBSEL) {
	(void) walk (e->e.subsel->result_list, cr_store);
	(void) walk (e->e.subsel->where_clause, cr_store);
    }
    return (WALK);
}

/* cr_store --- for certain operators, store 'em */

static int
cr_store (enode_t *e)
{
    void cr_add (char *);

    if (e->op == O_COL)
	cr_add (e->e.lit_name);
	
    return (WALK);
}

/* cr_add --- add this table/column to the list */

static void
cr_add (char *name)
{
    cross_ref_t *c;
    char *p, *table_name, *column_name;

    if ((p = strchr (name, '.')) != NULL) {
	*p = 0;
	table_name = strdup (name);
	column_name = strdup (p+1);
	*p = '.';
    }
    else {
	table_name = strdup ("*** AMBIGUOUS ***");
	column_name = strdup (name);
    }
    fprintf (fp, "%-20.20s %-20.20s %-30s %4ld %s\n",
      table_name,
      column_name,
      file_name,
      lineno,
      dotype (type)
    );
    free (table_name);
    free (column_name);

# ifdef COMMENT
    c = cr_new (name);
    c->file_name = file_name;
    c->lineno = lineno;
    c->type = type;
# endif /* COMMENT */
}

/* cr_new --- return a new cross_ref_t */

static cross_ref_t *
cr_new (char *name)
{
    int i;
    char *p;
	
    if (Ncross_ref >= MAXcross_ref) {
	MAXcross_ref += 10;
	if (Cross_ref == NULL)
	    Cross_ref = emalloc (MAXcross_ref * sizeof (Cross_ref[0]));
	else
	    Cross_ref = erealloc (Cross_ref,
	      MAXcross_ref * sizeof (Cross_ref[0]));
    }
	
    i = Ncross_ref++;
    Cross_ref[i] = emalloc (sizeof (cross_ref_t));
    memset (Cross_ref[i], 0, sizeof (cross_ref_t));
    Cross_ref[i]->name = strdup (name);
    if ((p = strchr (name, '.'))) {
	*p = 0;
	Cross_ref[i]->table_name = strdup (name);
	Cross_ref[i]->column_name = strdup (p+1);
	*p = '.';
    }
    else {
	Cross_ref[i]->column_name = strdup (name);
	/* this is overkill, but I'll free it just after I print it */
	Cross_ref[i]->table_name = strdup ("*** AMBIGUOUS ***");
    }
    return (Cross_ref[i]);
}

/* cr_cmp --- compare two elements for qsort */

static int
cr_cmp (cross_ref_t **a, cross_ref_t **b)
{
    int cmp = 0;

    if ((*a)->table_name && (*b)->table_name)
        cmp = nstrcmp ((*a)->table_name, (*b)->table_name);
    if (cmp)
	return (cmp);
    cmp = 0;
    if ((*a)->column_name && (*b)->column_name)
        cmp = nstrcmp ((*a)->column_name, (*b)->column_name);
    if (cmp)
	return (cmp);
    cmp = (*b)->type - (*a)->type;
    if (cmp)
	return (cmp);
    cmp = 0;
    if ((*a)->file_name && (*b)->file_name)
        cmp = nstrcmp ((*a)->file_name, (*b)->file_name);
    if (cmp)
	return (cmp);
    return ((*a)->lineno - (*b)->lineno);
}

/* dotype --- return a printable version of type */

static char *
dotype (int type)
{
    switch (type) {
    case DELETE:
	return ("DELETE");
	break;
    case INSERT:
	return ("INSERT");
	break;
    case SELECT:
	return ("SELECT");
	break;
    case UPDATE:
	return ("UPDATE");
	break;
    case WHERE:
	return ("WHERE");
	break;
    default:
	return ("I don't know");
	break;
    }
}

/* nstrcmp --- strcmp that can handle NULL pointers */

static int
nstrcmp (register char *a, register char *b)
{
    if (a == NULL && b == NULL)
	return (0);
    if (a == NULL && b != NULL)
	return (1);
    if (a != NULL && b == NULL)
	return (-1);
    if (a < 200 || b < 200) {
	fprintf (stderr, "nstrcmp: a = %x, b = %x\n", a, b);
	if (a >= 200)
		fprintf (stderr, "a = '%s'\n", a);
	if (b >= 200)
		fprintf (stderr, "b = '%s'\n", b);
	return (0);
    }
    return (strcmp (a, b));
}
